import os
import torch
from torch.utils.ffi import create_extension

this_file = os.path.dirname(__file__)

sources = []
headers = []
defines = []
with_cuda = False

if torch.cuda.is_available():
    print('Including CUDA code.')
    sources += ['pytorch_fft/src/th_fft_cuda.c']
    headers += ['pytorch_fft/src/th_fft_cuda.h']
    defines += [('WITH_CUDA', None)]
    with_cuda = True

ffi = create_extension(
    'pytorch_fft._ext.th_fft',
    package=True,
    headers=headers,
    sources=sources,
    define_macros=defines,
    relative_to=__file__,
    with_cuda=with_cuda,
    include_dirs=[os.getcwd() + '/pytorch_fft/src'],
    library_dirs=['/usr/local/cuda/lib64'], 
    libraries=['cufft']
)

if __name__ == '__main__':
    ffi.build()



import torch
from .fft import fft,ifft,fft2,ifft2,fft3,ifft3


class Fft(torch.autograd.Function):
    def forward(self, X_re, X_im):
        return fft(X_re, X_im)

    def backward(self, grad_output_re, grad_output_im):
        N = grad_output_re.size(-1)
        gr, gi = ifft(grad_output_re,-grad_output_im)
        gr, gi = gr * N, -gi * N
        return gr,gi


class Ifft(torch.autograd.Function):

    def forward(self, k_re, k_im):
        return ifft(k_re, k_im)

    def backward(self, grad_output_re, grad_output_im):
        gr, gi = fft(grad_output_re,-grad_output_im)
        gr, gi = gr, -gi
        return gr, gi


class Fft2d(torch.autograd.Function):
    def forward(self, X_re, X_im):
        return fft2(X_re, X_im)

    def backward(self, grad_output_re, grad_output_im):
        N = grad_output_re.size(-1) * grad_output_re.size(-2)
        gr, gi = ifft2(grad_output_re,-grad_output_im)
        gr, gi = gr * N, -gi * N
        return gr,gi


class Ifft2d(torch.autograd.Function):

    def forward(self, k_re, k_im):
        return ifft2(k_re, k_im)

    def backward(self, grad_output_re, grad_output_im):
        gr, gi = fft2(grad_output_re,-grad_output_im)
        gr, gi = gr, -gi
        return gr, gi


class Fft3d(torch.autograd.Function):
    def forward(self, X_re, X_im):
        return fft3(X_re, X_im)

    def backward(self, grad_output_re, grad_output_im):
        N = grad_output_re.size(-1) * grad_output_re.size(-2) * grad_output_re.size(-3)
        gr, gi = ifft3(grad_output_re,-grad_output_im)
        gr, gi = gr * N, -gi * N
        return gr,gi


class Ifft3d(torch.autograd.Function):

    def forward(self, k_re, k_im):
        return ifft3(k_re, k_im)

    def backward(self, grad_output_re, grad_output_im):
        gr, gi = fft3(grad_output_re,-grad_output_im)
        gr, gi = gr, -gi
        return gr, gi



# functions/fft.py
import torch
from .._ext import th_fft

def _fft(X_re, X_im, f, rank):
    if not(X_re.dim() >= rank+1 and X_im.dim() >= rank+1): 
        raise ValueError("Inputs must have at least {} dimensions.".format(rank+1))
    if not(X_re.is_cuda and X_im.is_cuda): 
        raise ValueError("Input must be a CUDA tensor.")
    if not(X_re.is_contiguous() and X_im.is_contiguous()):
        raise ValueError("Input must be contiguous.")

    Y1, Y2 = tuple(X_re.new(*X_re.size()).zero_() for _ in range(2))
    f(X_re, X_im, Y1, Y2)
    return (Y1, Y2)

def fft(X_re, X_im): 
    if 'Float' in type(X_re).__name__ :
        f = th_fft.th_Float_fft1
    elif 'Double' in type(X_re).__name__: 
        f = th_fft.th_Double_fft1
    else: 
        raise NotImplementedError
    return _fft(X_re, X_im, f, 1)

def ifft(X_re, X_im): 
    N = X_re.size(-1)
    if 'Float' in type(X_re).__name__ :
        f = th_fft.th_Float_ifft1
    elif 'Double' in type(X_re).__name__: 
        f = th_fft.th_Double_ifft1
    else: 
        raise NotImplementedError   
    Y1, Y2 = _fft(X_re, X_im, f, 1)
    return (Y1/N, Y2/N)

def fft2(X_re, X_im): 
    if 'Float' in type(X_re).__name__ :
        f = th_fft.th_Float_fft2
    elif 'Double' in type(X_re).__name__: 
        f = th_fft.th_Double_fft2
    else: 
        raise NotImplementedError
    return _fft(X_re, X_im, f, 2)

def ifft2(X_re, X_im): 
    N = X_re.size(-1)*X_re.size(-2)
    if 'Float' in type(X_re).__name__ :
        f = th_fft.th_Float_ifft2
    elif 'Double' in type(X_re).__name__: 
        f = th_fft.th_Double_ifft2
    else: 
        raise NotImplementedError   
    Y1, Y2 = _fft(X_re, X_im, f, 2)
    return (Y1/N, Y2/N)

def fft3(X_re, X_im): 
    if 'Float' in type(X_re).__name__ :
        f = th_fft.th_Float_fft3
    elif 'Double' in type(X_re).__name__: 
        f = th_fft.th_Double_fft3
    else: 
        raise NotImplementedError
    return _fft(X_re, X_im, f, 3)

def ifft3(X_re, X_im): 
    N = X_re.size(-1)*X_re.size(-2)*X_re.size(-3)
    if 'Float' in type(X_re).__name__ :
        f = th_fft.th_Float_ifft3
    elif 'Double' in type(X_re).__name__: 
        f = th_fft.th_Double_ifft3
    else: 
        raise NotImplementedError   
    Y1, Y2 = _fft(X_re, X_im, f, 3)
    return (Y1/N, Y2/N)

_s = slice(None, None, None)

def _rfft(X, f, rank):
    if not(X.dim() >= rank+1): 
        raise ValueError("Input must have at least {} dimensions.".format(rank+1))
    if not(X.is_cuda): 
        raise ValueError("Input must be a CUDA tensor.")
    if not(X.is_contiguous()):
        raise ValueError("Input must be contiguous.")

    new_size = tuple(X.size())[:-1] + (X.size(-1)//2 + 1,)
    # new_size = tuple(X.size())
    Y1, Y2 = tuple(X.new(*new_size).zero_() for _ in range(2))
    f(X, Y1, Y2)
    # i = tuple(_s for _ in range(X.dim()-1)) + (slice(None, X.size(-1)//2 + 1, ),)
    # print(Y1, i)
    # return (Y1[i], Y2[i])
    return (Y1, Y2)

def rfft(X): 
    if 'Float' in type(X).__name__ :
        f = th_fft.th_Float_rfft1
    elif 'Double' in type(X).__name__: 
        f = th_fft.th_Double_rfft1
    else: 
        raise NotImplementedError
    return _rfft(X, f, 1)

def rfft2(X): 
    if 'Float' in type(X).__name__ :
        f = th_fft.th_Float_rfft2
    elif 'Double' in type(X).__name__: 
        f = th_fft.th_Double_rfft2
    else: 
        raise NotImplementedError
    return _rfft(X, f, 2)

def rfft3(X): 
    if 'Float' in type(X).__name__ :
        f = th_fft.th_Float_rfft3
    elif 'Double' in type(X).__name__: 
        f = th_fft.th_Double_rfft3
    else: 
        raise NotImplementedError
    return _rfft(X, f, 3)

def _irfft(X_re, X_im, f, rank):
    if not(X_re.dim() >= rank+1 and X_im.dim() >= rank+1): 
        raise ValueError("Inputs must have at least {} dimensions.".format(rank+1))
    if not(X_re.is_cuda and X_im.is_cuda): 
        raise ValueError("Input must be a CUDA tensor.")
    if not(X_re.is_contiguous() and X_im.is_contiguous()):
        raise ValueError("Input must be contiguous.")

    # raise NotImplementedError
    N = (X_re.size(-1) - 1)*2
    new_size = tuple(X_re.size())[:-1] + (N,)
    Y = X_re.new(*new_size).zero_()
    f(X_re, X_im, Y)
    M = 1
    for i in range(rank): 
        M *= new_size[-(i+1)]
    return Y/M

def irfft(X_re, X_im): 
    if 'Float' in type(X_re).__name__ :
        f = th_fft.th_Float_irfft1
    elif 'Double' in type(X_re).__name__: 
        f = th_fft.th_Double_irfft1
    else: 
        raise NotImplementedError
    return _irfft(X_re, X_im, f, 1)

def irfft2(X_re, X_im): 
    if 'Float' in type(X_re).__name__ :
        f = th_fft.th_Float_irfft2
    elif 'Double' in type(X_re).__name__: 
        f = th_fft.th_Double_irfft2
    else: 
        raise NotImplementedError
    return _irfft(X_re, X_im, f, 2)

def irfft3(X_re, X_im): 
    if 'Float' in type(X_re).__name__ :
        f = th_fft.th_Float_irfft3
    elif 'Double' in type(X_re).__name__: 
        f = th_fft.th_Double_irfft3
    else: 
        raise NotImplementedError
    return _irfft(X_re, X_im, f, 3)


import os
import sys

from setuptools import setup, find_packages

import build

this_file = os.path.dirname(__file__)

setup(
    name="pytorch_fft",
    version="0.9",
    description="A PyTorch wrapper for CUDA FFTs",
    url="https://github.com/locuslab/pytorch_fft",
    author="Eric Wong",
    author_email="ericwong@cs.cmu.edu",
    # Require cffi.
    install_requires=["cffi>=1.0.0"],
    setup_requires=["cffi>=1.0.0"],
    # Exclude the build files.
    packages=find_packages(exclude=["build"]),
    # Package where to put the extensions. Has to be a prefix of build.py.
    ext_package="",
    # Extensions to compile.
    cffi_modules=[
        os.path.join(this_file, "build.py:ffi")
    ],
)